/*
 *    LedControl.cpp - A library for controling Leds with a MAX7219/MAX7221
 *    Copyright (c) 2007 Eberhard Fahle
 *
 *    Permission is hereby granted, free of charge, to any person
 *    obtaining a copy of this software and associated documentation
 *    files (the "Software"), to deal in the Software without
 *    restriction, including without limitation the rights to use,
 *    copy, modify, merge, publish, distribute, sublicense, and/or sell
 *    copies of the Software, and to permit persons to whom the
 *    Software is furnished to do so, subject to the following
 *    conditions:
 *
 *    This permission notice shall be included in all copies or
 *    substantial portions of the Software.
 *
 *    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 *    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 *    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 *    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 *    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 *    OTHER DEALINGS IN THE SOFTWARE.
 */


#include "LedControl.h"

//the opcodes for the MAX7221 and MAX7219
#define OP_NOOP   0
#define OP_DIGIT0 1
#define OP_DIGIT1 2
#define OP_DIGIT2 3
#define OP_DIGIT3 4
#define OP_DIGIT4 5
#define OP_DIGIT5 6
#define OP_DIGIT6 7
#define OP_DIGIT7 8
#define OP_DECODEMODE  9
#define OP_INTENSITY   10
#define OP_SCANLIMIT   11
#define OP_SHUTDOWN    12
#define OP_DISPLAYTEST 15


LedControl::LedControl(int dataPin, int clkPin, int csPin, int numDevices) {
    SPI_MOSI=dataPin;
    SPI_CLK=clkPin;
    SPI_CS=csPin;
    if(numDevices<=0 || numDevices>8 )
	numDevices=8;
    maxDevices=numDevices;
    pinMode(SPI_MOSI,OUTPUT);
    pinMode(SPI_CLK,OUTPUT);
    pinMode(SPI_CS,OUTPUT);
    digitalWrite(SPI_CS,HIGH);
    SPI_MOSI=dataPin;
    for(int i=0;i<64;i++)
	status[i]=0x00;
    for(int i=0;i<maxDevices;i++) {
	spiTransfer(i,OP_DISPLAYTEST,0);
	//scanlimit is set to max on startup
	setScanLimit(i,7);
	//decode is done in source
	spiTransfer(i,OP_DECODEMODE,0);
	clearDisplay(i);
	//we go into shutdown-mode on startup
	shutdown(i,true);
    }
}

int LedControl::getDeviceCount() {
    return maxDevices;
}

void LedControl::shutdown(int addr, bool b) {
    if(addr<0 || addr>=maxDevices)
	return;
    if(b)
	spiTransfer(addr, OP_SHUTDOWN,0);
    else
	spiTransfer(addr, OP_SHUTDOWN,1);
}

void LedControl::setScanLimit(int addr, int limit) {
    if(addr<0 || addr>=maxDevices)
	return;
    if(limit>=0 || limit<8)
    	spiTransfer(addr, OP_SCANLIMIT,limit);
}

void LedControl::setIntensity(int addr, int intensity) {
    if(addr<0 || addr>=maxDevices)
	return;
    if(intensity>=0 || intensity<16)
	spiTransfer(addr, OP_INTENSITY,intensity);

}

void LedControl::clearDisplay(int addr) {
    int offset;

    if(addr<0 || addr>=maxDevices)
	return;
    offset=addr*8;
    for(int i=0;i<8;i++) {
	status[offset+i]=0;
	spiTransfer(addr, i+1,status[offset+i]);
    }
}

void LedControl::setLed(int addr, int row, int column, boolean state) {
    int offset;
    byte val=0x00;

    if(addr<0 || addr>=maxDevices)
	return;
    if(row<0 || row>7 || column<0 || column>7)
	return;
    offset=addr*8;
    val=B10000000 >> column;
    if(state)
	status[offset+row]=status[offset+row]|val;
    else {
	val=~val;
	status[offset+row]=status[offset+row]&val;
    }
    spiTransfer(addr, row+1,status[offset+row]);
}

void LedControl::setRow(int addr, int row, byte value) {
    int offset;
    if(addr<0 || addr>=maxDevices)
	return;
    if(row<0 || row>7)
	return;
    offset=addr*8;
    status[offset+row]=value;
    spiTransfer(addr, row+1,status[offset+row]);
}

void LedControl::setColumn(int addr, int col, byte value) {
    byte val;

    if(addr<0 || addr>=maxDevices)
	return;
    if(col<0 || col>7)
	return;
    for(int row=0;row<8;row++) {
	val=value >> (7-row);
	val=val & 0x01;
	setLed(addr,row,col,val);
    }
}

void LedControl::setDigit(int addr, int digit, byte value, boolean dp) {
    int offset;
    byte v;

    if(addr<0 || addr>=maxDevices)
	return;
    if(digit<0 || digit>7 || value>15)
	return;
    offset=addr*8;
    v=charTable[value];
    if(dp)
	v|=B10000000;
    status[offset+digit]=v;
    spiTransfer(addr, digit+1,v);

}

void LedControl::setChar(int addr, int digit, char value, boolean dp) {
    int offset;
    byte index,v;

    if(addr<0 || addr>=maxDevices)
	return;
    if(digit<0 || digit>7)
 	return;
    offset=addr*8;
    index=(byte)value;
    if(index >127) {
	//nothing define we use the space char
	value=32;
    }
    v=charTable[index];
    if(dp)
	v|=B10000000;
    status[offset+digit]=v;
    spiTransfer(addr, digit+1,v);
}

void LedControl::spiTransfer(int addr, volatile byte opcode, volatile byte data) {
    //Create an array with the data to shift out
    int offset=addr*2;
    int maxbytes=maxDevices*2;

    for(int i=0;i<maxbytes;i++)
	spidata[i]=(byte)0;
    //put our device data into the array
    spidata[offset+1]=opcode;
    spidata[offset]=data;
    //enable the line
    digitalWrite(SPI_CS,LOW);
    //Now shift out the data
    for(int i=maxbytes;i>0;i--)
 	shiftOut(SPI_MOSI,SPI_CLK,MSBFIRST,spidata[i-1]);
    //latch the data onto the display
    digitalWrite(SPI_CS,HIGH);
}

void LedControl::singledigit(double n, int p)
{
 if(p == 0){
  Pos=0;}
if(p == 1){
  Pos=5;}
if(n==0){setLed(0,0,(1) + (Pos),true);setLed(0,1,(0) + (Pos),true);setLed(0,2,(0) + (Pos),true);setLed(0,3,(0) + (Pos),true);setLed(0,4,(1) + (Pos),true);setLed(0,1,(2) + (Pos),true);setLed(0,2,(2) + (Pos),true);setLed(0,3,(2) + (Pos),true);}
if(n==1){setLed(0,0,(2) + (Pos),true);setLed(0,1,(2) + (Pos),true);setLed(0,2,(2) + (Pos),true);setLed(0,3,(2) + (Pos),true);setLed(0,4,(2) + (Pos),true);}
if(n==2){setLed(0,0,(1) + (Pos),true);setLed(0,1,(0) + (Pos),true);setLed(0,1,(2) + (Pos),true);setLed(0,2,(2) + (Pos),true);setLed(0,3,(1) + (Pos),true);setLed(0,4,(0) + (Pos),true);setLed(0,4,(1) + (Pos),true);setLed(0,4,(2) + (Pos),true);}
if(n==3){setLed(0,0,(0) + (Pos),true);setLed(0,0,(1) + (Pos),true);setLed(0,1,(2) + (Pos),true);setLed(0,2,(1) + (Pos),true);setLed(0,3,(2) + (Pos),true);setLed(0,4,(0) + (Pos),true);setLed(0,4,(1) + (Pos),true);}
if(n==4){setLed(0,0,(2) + (Pos),true);setLed(0,1,(1) + (Pos),true);setLed(0,1,(2) + (Pos),true);setLed(0,2,(0) + (Pos),true);setLed(0,2,(2) + (Pos),true);setLed(0,3,(0) + (Pos),true);setLed(0,3,(1) + (Pos),true);setLed(0,3,(2) + (Pos),true);setLed(0,4,(2) + (Pos),true);}
if(n==5){setLed(0,0,(0) + (Pos),true);setLed(0,0,(1) + (Pos),true);setLed(0,0,(2) + (Pos),true);setLed(0,1,(0) + (Pos),true);setLed(0,2,(0) + (Pos),true);setLed(0,2,(1) + (Pos),true);setLed(0,3,(2) + (Pos),true);setLed(0,4,(0) + (Pos),true);setLed(0,4,(1) + (Pos),true);}
if(n==6){setLed(0,0,(1) + (Pos),true);setLed(0,0,(2) + (Pos),true);setLed(0,1,(0) + (Pos),true);setLed(0,2,(0) + (Pos),true);setLed(0,2,(1) + (Pos),true);setLed(0,2,(2) + (Pos),true);setLed(0,3,(0) + (Pos),true);setLed(0,3,(2) + (Pos),true);setLed(0,4,(1) + (Pos),true);}
if(n==7){setLed(0,0,(0) + (Pos),true);setLed(0,0,(1) + (Pos),true);setLed(0,0,(2) + (Pos),true);setLed(0,0,(2) + (Pos),true);setLed(0,1,(2) + (Pos),true);setLed(0,2,(2) + (Pos),true);setLed(0,3,(2) + (Pos),true);setLed(0,4,(2) + (Pos),true);}
if(n==8){setLed(0,0,(1) + (Pos),true);setLed(0,1,(0) + (Pos),true);setLed(0,1,(2) + (Pos),true);setLed(0,2,(1) + (Pos),true);setLed(0,3,(0) + (Pos),true);setLed(0,3,(2) + (Pos),true);setLed(0,4,(1) + (Pos),true);setLed(0,0,(0) + (Pos),true);setLed(0,0,(2) + (Pos),true);setLed(0,2,(0) + (Pos),true);setLed(0,2,(2) + (Pos),true);setLed(0,4,(0) + (Pos),true);setLed(0,4,(2) + (Pos),true);}
if(n==9){setLed(0,0,(0) + (Pos),true);setLed(0,0,(1) + (Pos),true);setLed(0,0,(2) + (Pos),true);setLed(0,1,(0) + (Pos),true);setLed(0,1,(2) + (Pos),true);setLed(0,2,(0) + (Pos),true);setLed(0,2,(1) + (Pos),true);setLed(0,2,(2) + (Pos),true);setLed(0,3,(2) + (Pos),true);setLed(0,4,(2) + (Pos),true);}
}


void LedControl::multipledigit(double n){
n=round(n);
singledigit(fmod(n,10),1);
if(fmod(n/10,10) >= 1){
singledigit(fmod(floor(n/10),10),0);}
if(n/100>=1){
  for(int i=0;i<8;i++){setRow(0,i,errorsymbol[i]);}
}
}

void LedControl::multipledigitscroll(double n){
n=round(n);

if(n/10000>=1){
  for(int i=0;i<8;i++){setRow(0,i,errorsymbol[i]);}
}
else
{
if(n/1000>=1){
if(floor(n/1000)==1){scrollMessage(scrollNo1);}
if(floor(n/1000)==2){scrollMessage(scrollNo2);}
if(floor(n/1000)==3){scrollMessage(scrollNo3);}
if(floor(n/1000)==4){scrollMessage(scrollNo4);}
if(floor(n/1000)==5){scrollMessage(scrollNo5);}
if(floor(n/1000)==6){scrollMessage(scrollNo6);}
if(floor(n/1000)==7){scrollMessage(scrollNo7);}
if(floor(n/1000)==8){scrollMessage(scrollNo8);}
if(floor(n/1000)==9){scrollMessage(scrollNo9);;}
scrollMessage(comma);
}

if(n/100 >= 1){
if(fmod(floor(n/100),10)==0){scrollMessage(scrollNo0);}
if(fmod(floor(n/100),10)==1){scrollMessage(scrollNo1);}
if(fmod(floor(n/100),10)==2){scrollMessage(scrollNo2);}
if(fmod(floor(n/100),10)==3){scrollMessage(scrollNo3);}
if(fmod(floor(n/100),10)==4){scrollMessage(scrollNo4);}
if(fmod(floor(n/100),10)==5){scrollMessage(scrollNo5);}
if(fmod(floor(n/100),10)==6){scrollMessage(scrollNo6);}
if(fmod(floor(n/100),10)==7){scrollMessage(scrollNo7);}
if(fmod(floor(n/100),10)==8){scrollMessage(scrollNo8);}
if(fmod(floor(n/100),10)==9){scrollMessage(scrollNo9);}
}
if(n/10 >= 1){
  if(fmod(floor(n/10),10)==0){scrollMessage(scrollNo0);}
  if(fmod(floor(n/10),10)==1){scrollMessage(scrollNo1);}
  if(fmod(floor(n/10),10)==2){scrollMessage(scrollNo2);}
  if(fmod(floor(n/10),10)==3){scrollMessage(scrollNo3);}
  if(fmod(floor(n/10),10)==4){scrollMessage(scrollNo4);}
  if(fmod(floor(n/10),10)==5){scrollMessage(scrollNo5);}
  if(fmod(floor(n/10),10)==6){scrollMessage(scrollNo6);}
  if(fmod(floor(n/10),10)==7){scrollMessage(scrollNo7);}
  if(fmod(floor(n/10),10)==8){scrollMessage(scrollNo8);}
  if(fmod(floor(n/10),10)==9){scrollMessage(scrollNo9);}
}
if(n/1 >= 1){
  if(fmod(n,10) == 0){scrollMessage(scrollNo0);}
  if(fmod(n,10) == 1){scrollMessage(scrollNo1);}
  if(fmod(n,10) == 2){scrollMessage(scrollNo2);}
  if(fmod(n,10) == 3){scrollMessage(scrollNo3);}
  if(fmod(n,10) == 4){scrollMessage(scrollNo4);}
  if(fmod(n,10) == 5){scrollMessage(scrollNo5);}
  if(fmod(n,10) == 6){scrollMessage(scrollNo6);}
  if(fmod(n,10) == 7){scrollMessage(scrollNo7);}
  if(fmod(n,10) == 8){scrollMessage(scrollNo8);}
  if(fmod(n,10) == 9){scrollMessage(scrollNo9);}
}
if(n==0){
  scrollMessage(scrollNo0);
}
  scrollMessage(space);
  scrollMessage(space);
}
}

void LedControl::scrollFont() {
    for (int counter=0x20;counter<0x80;counter++){
        loadBufferLong(counter);
        delay(500);
    }
}

// Scroll Message
void LedControl::scrollMessage(const unsigned char * messageString) {
    int counter = 0;
    int myChar=0;
    do {
        // read back a char
        myChar =  pgm_read_byte_near(messageString + counter);
        if (myChar != 0){
            loadBufferLong(myChar);
        }
        counter++;
    }
    while (myChar != 0);
}
// Load character into scroll buffer
void LedControl::loadBufferLong(int ascii){
    if (ascii >= 0x20 && ascii <=0x7f){
        for (int a=0;a<7;a++){                      // Loop 7 times for a 5x7 font
            unsigned long c = pgm_read_byte_near(font5x7 + ((ascii - 0x20) * 8) + a);     // Index into character table to get row data
            unsigned long x = bufferLong [a*2];     // Load current scroll buffer
            x = x | c;                              // OR the new character onto end of current
            bufferLong [a*2] = x;                   // Store in buffer
        }
        byte count = pgm_read_byte_near(font5x7 +((ascii - 0x20) * 8) + 7);     // Index into character table for kerning data
        for (byte x=0; x<count;x++){
            rotateBufferLong();
            printBufferLong();
            delay(scrollDelay);
        }
    }
}
// Rotate the buffer
void LedControl::rotateBufferLong(){
    for (int a=0;a<7;a++){                      // Loop 7 times for a 5x7 font
        unsigned long x = bufferLong [a*2];     // Get low buffer entry
        byte b = bitRead(x,31);                 // Copy high order bit that gets lost in rotation
        x = x<<1;                               // Rotate left one bit
        bufferLong [a*2] = x;                   // Store new low buffer
        x = bufferLong [a*2+1];                 // Get high buffer entry
        x = x<<1;                               // Rotate left one bit
        bitWrite(x,0,b);                        // Store saved bit
        bufferLong [a*2+1] = x;                 // Store new high buffer
    }
}
// Display Buffer on LED matrix
void LedControl::printBufferLong(){
  for (int a=0;a<7;a++){                    // Loop 7 times for a 5x7 font
    unsigned long x = bufferLong [a*2+1];   // Get high buffer entry
    byte y = x;                             // Mask off first character
    setRow(3,a,y);                       // Send row to relevent MAX7219 chip
    x = bufferLong [a*2];                   // Get low buffer entry
    y = (x>>24);                            // Mask off second character
    setRow(2,a,y);                       // Send row to relevent MAX7219 chip
    y = (x>>16);                            // Mask off third character
    setRow(1,a,y);                       // Send row to relevent MAX7219 chip
    y = (x>>8);                             // Mask off forth character
    setRow(0,a,y);                       // Send row to relevent MAX7219 chip
  }
}
